use std::{
    future::Future,
    pin::Pin,
    task::{Context, Poll},
};

use pin_project::pin_project;

trait More {
    type Trait: ?Sized + Trait;
}

trait Member {
    type Trait: ?Sized + Trait;
    type Associated<I: ?Sized + Impl<Self::Trait>>: ?Sized + Trait;
}

trait SizedMember {
    type Trait: ?Sized + Trait;
}

impl<T: ?Sized + Member> SizedMember for T {
    type Trait = T::Trait;
}

trait Trait {
    type More: ?Sized + More;
    type Member: ?Sized + Member<Trait = Self>;
    type Sized: ?Sized + SizedMember<Trait = Self>;
}

trait ImplMember<T: ?Sized + Member> {
    type Associated: Impl<T::Associated<Self>>
    where
        Self: Impl<T::Trait>;
}

trait ImplSizedMember<T: ?Sized + SizedMember> {}

impl<T: ?Sized + Member, I: ?Sized + ImplMember<T>> ImplSizedMember<T> for I {}

trait ImplBase<T: ?Sized + Trait>: ImplMember<T::Member> + ImplSizedMember<T::Sized> {
    type More: ?Sized + ImplMore<T::More>;
}

trait Impl<T: ?Sized + Trait>: ImplBase<T, More = Self> {}

impl<T: ?Sized + Trait, I: ?Sized + ImplBase<T, More = Self>> Impl<T> for I {}

trait ImplMore<T: ?Sized + More>: Impl<T::Trait> {}

enum Empty {}

impl More for Empty {
    type Trait = Self;
}

impl Member for Empty {
    type Trait = Self;
    type Associated<I: ?Sized> = Self;
}

impl Trait for Empty {
    type More = Self;
    type Member = Self;
    type Sized = Self;
}

impl<T: ?Sized> ImplMore<Empty> for T {}

impl<T: ?Sized> ImplMember<Empty> for T {
    type Associated = Empty
    where
        Self: Impl<Empty>;
}

impl<T: ?Sized> ImplBase<Empty> for T {
    type More = Self;
}

struct Futures;

impl Member for Futures {
    type Trait = Self;
    type Associated<I: ?Sized + Impl<Self::Trait>> = Empty;
}

impl Trait for Futures {
    type More = Empty;
    type Member = Self;
    type Sized = Self;
}

impl<T: ?Sized + Future> ImplMember<Futures> for T {
    type Associated = T::Output
    where
        Self: Impl<<Futures as Member>::Trait>;
}

impl<T: ?Sized + Future> ImplBase<Futures> for T {
    type More = Self;
}

struct No<T: ?Sized>(T);

impl<T: ?Sized + Trait> Member for No<T> {
    type Trait = T;
    type Associated<I: ?Sized + Impl<Self::Trait>> = Empty;
}

impl<T: ?Sized + Trait, I: ?Sized> ImplMember<No<T>> for I {
    type Associated = Empty
    where
        Self: Impl<<No<T> as Member>::Trait>;
}

struct Is<A>(A);

impl<A> SizedMember for Is<A> {
    type Trait = Self;
}

impl<A> Trait for Is<A> {
    type More = Empty;
    type Member = No<Self>;
    type Sized = Self;
}

impl<I> ImplSizedMember<Is<I>> for I {}

impl<I> ImplBase<Is<I>> for I {
    type More = Self;
}

struct FutureIs<A>(A);

impl<A> More for FutureIs<A> {
    type Trait = Futures;
}

impl<A> Trait for FutureIs<A> {
    type More = Self;
    type Member = No<Self>;
    type Sized = No<Self>;
}

impl<F: ?Sized + Future> ImplBase<FutureIs<F::Output>> for F {
    type More = Self;
}

impl<F: ?Sized + Future> ImplMore<FutureIs<F::Output>> for F {}

#[pin_project]
struct TraitFuture<F: ?Sized + Impl<Futures>>(#[pin] F);

impl<F: ?Sized + Impl<Futures>> Future for TraitFuture<F> {
    type Output = F::Associated;

    fn poll(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Self::Output> {
        todo!()
    }
}

trait Functor {
    type Trait<A>: ?Sized + Trait;
    fn pure<A>(a: A) -> impl Impl<Self::Trait<A>>;
    fn map<A, B>(a: impl Impl<Self::Trait<A>>, f: impl FnOnce(A) -> B)
        -> impl Impl<Self::Trait<B>>;
}

impl Functor for Futures {
    type Trait<A> = FutureIs<A>;

    fn pure<A>(a: A) -> impl Impl<Self::Trait<A>> {
        async { a }
    }

    fn map<A, B>(
        a: impl Impl<Self::Trait<A>>,
        f: impl FnOnce(A) -> B,
    ) -> impl Impl<Self::Trait<B>> {
        let fut = TraitFuture(upcast(a));
        async {
            let a = fut.await;
            f(todo!())
        }
    }
}

fn upcast_more<T: ?Sized + Trait, I: ImplBase<T>>(value: I::More) -> impl ImplMore<T::More>
where
    <I as ImplBase<T>>::More: std::marker::Sized,
{
    value
}

fn upcast<T: ?Sized + Trait, I: Impl<T>>(value: I) -> impl ImplMore<T::More> {
    upcast_more::<T, I>(value)
}
